# coding:utf-8
import os


# 建模代码目录
ROOT = os.getcwd()
# CSV 文件总目录
csvRoot = ROOT + '/data/'
# 单个项目 CSV 目录
csvFile= csvRoot


# 准备入模的数据，已经完成清洗
makeModelDataFillnaFile = 'model_data_xgb_201912_lianjia_30'
ModelDataFillnaFile = 'only_model_data'


# 分箱 好坏区间及占比 IV
ruleFilePri = 'dt_rule_all'
ruleModelFilePri = 'dt_rule_model'
ruleModelFile = 'reg_rule_model'

OutModelFile = 'model_'

# 入模原始数据
scoreCardDataFile = 'model_origin_data_'
# 模型变量相关性
comParamCorrFile = 'com_param_corr_'
# 模型样本最终评分
checkDataFile = 'model_score_'
# 评分卡参数文件
scoreCardParamFile = 'score_card_param_'
# 评分卡pkl文件
scoreCardPklFile = 'score_card_pkl_'
# woe明细文件
woeFile = 'woe_'
# 是否输出woe明细文件
woe_file_output = False

# id 和 y 的变量命名
# col_head = ['cus_num', 'y']
col_head = ['employee_no', 'bad']




# coding:utf-8

import numpy as np
import matplotlib.pyplot as plot
import pandas as pd

from sklearn.metrics import roc_curve

from lapras.ModelCode.utils import cost
from lapras.ModelCode.utils import train


from lapras.ModelCode.utils.FileUtil import write_data
from lapras.DealConfig import OutModelFileName


def ks_calc_cross(pred, y_label):
    '''
    功能: 计算KS值，输出对应分割点和累计分布函数曲线图
    输入值:
    data: 二维数组或dataframe，包括模型得分和真实的标签
    pred: 一维数组或series，代表模型得分（一般为预测正类的概率）
    y_label: 一维数组或series，代表真实的标签（{0,1}或{-1,1}）
    输出值:
    'ks': KS值，'crossdens': 好坏客户累积概率分布以及其差值gap
    '''
    crossfreq = pd.crosstab(pred, y_label)
    crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
    crossdens['gap'] = abs(crossdens[0] - crossdens[1])
    ks = crossdens[crossdens['gap'] == crossdens['gap'].max()]
    return float(ks['gap'])


def ks_calc_auc(pred, y_label):
    '''
    功能: 计算KS值，输出对应分割点和累计分布函数曲线图
    输入值:
    data: 二维数组或dataframe，包括模型得分和真实的标签
    pred: 一维数组或series，代表模型得分（一般为预测正类的概率）
    y_label: 一维数组或series，代表真实的标签（{0,1}或{-1,1}）
    输出值:
    'ks': KS值
    '''
    fpr, tpr, thresholds = roc_curve(y_label, pred)
    ks = max(tpr - fpr)
    return ks


def KS(testx, testy, theta, ks_bucket_num=10):
    tot_bad = np.sum(testy == 1)
    tot_good = np.sum(testy == 0)
    testx = np.column_stack((np.ones((testx.shape[0], 1), dtype=float), testx))
    n = theta.shape[0]
    h = cost.sigmoidFunction(testx, theta.reshape(n, 1)) # prob
    data = np.column_stack((h, testy))
    data = data[np.argsort(data[:, 0])]
    # np.random.shuffle(data)
    m = data.shape[0]
    sum_bad_percent = [0.0]
    sum_good_percent = [0.0]
    score = []
    for i in range(ks_bucket_num):
        if i == ks_bucket_num - 1:
            data_seg = data[int(m * i / ks_bucket_num):]
        else:
            data_seg = data[int(m * i / ks_bucket_num):int(m * (i + 1) / ks_bucket_num)]
        bad = np.sum(data_seg[:, 1] == 1)
        good = np.sum(data_seg[:, 1] == 0)
        if good == 0:
            percent = 999.0
        else:
            percent = 1.0 * bad / good
        score.append([percent, bad, good])
    score = np.array(score)
    score = score[np.argsort(score[:, 0])]
    # print score
    for i in range(ks_bucket_num):
        # i = 10 - j - 1
        bad = score[i, 1]
        good = score[i, 2]
        # print bad, good, score[i, 0]
        bad_percent = 1.0 * bad / tot_bad
        good_percent = 1.0 * good / tot_good
        sum_bad_percent.append(sum_bad_percent[-1] + bad_percent)
        sum_good_percent.append(sum_good_percent[-1] + good_percent)
    sum_bad_percent = np.array(sum_bad_percent)
    sum_good_percent = np.array(sum_good_percent)
    ks = np.max(sum_good_percent - sum_bad_percent)
    # print 'k-s = %f ' % (ks)
    return (ks, sum_bad_percent, sum_good_percent)


def KS_draw(sum_bad_percent, sum_good_percent,ks_bucket_num=10):
    index = np.array(range(ks_bucket_num + 1))
    plot.figure(0)
    plot.title(u'max K-S')
    plot.axis([0, ks_bucket_num, 0, 1])
    plot1, = plot.plot(index, sum_bad_percent, 'bo-')
    plot2, = plot.plot(index, sum_good_percent, 'ro-')
    # plot.plot([0, 10], [0, 1], '-')
    plot.legend([plot1, plot2], ['bad', 'good'])
    plot.grid(True)
    ax = plot.gca()
    ax.set_xticks(index)
    ax.set_xticklabels(('0%', '10%', '20%', '30%', '40%', '50%', '60%', '70%', '80%', '90%', '100%'))
    plot.show()
    # plot.savefig(img_path + '/ks_' + pageID + '.png')
    # plot.savefig('img/ks_'+pageID+'.png')


def ROC(testx, testy, theta):
    m = testx.shape[0]
    n = testx.shape[1] + 1
    testx = np.column_stack((np.ones((m, 1), dtype=float), testx))
    n = theta.shape[0]
    h = cost.sigmoidFunction(testx, theta.reshape(n, 1)).reshape(m, )
    TPR_num = []
    FNR_num = []
    tot_true = np.sum(testy == 1)
    tot_false = np.sum(testy == 0)
    alpha_list = np.linspace(0, 1, 1001)
    for alpha in alpha_list:
        TPR_num.append(np.sum(np.bitwise_and(testy == 1, h >= alpha)))
        FNR_num.append(np.sum(np.bitwise_and(testy == 0, h < alpha)))
    TPR_num = np.array(TPR_num)
    FNR_num = np.array(FNR_num)
    TPR = 1.0 * TPR_num / tot_true
    FNR = 1.0 - 1.0 * FNR_num / tot_false
    length = len(TPR)
    AUC = 0.0
    for i in range(length - 1):
        AUC += (TPR[i] + TPR[i + 1]) * (FNR[i] - FNR[i + 1]) / 2.0
    # print 'AUC = %f ' % AUC
    return (AUC, FNR, TPR)


def ROC_draw(FNR, TPR):
    plot.figure(1)
    plot.grid(True)
    plot.axis([0, 1, 0, 1])
    plot.title(u'max ROC')
    plot.plot(FNR, TPR, 'o-')
    plot.plot([0, 1], [0, 1], '--')
    plot.show()
    # plot.savefig(img_path + '/ROC_' + pageID + '.png')
    # plot.savefig('img/ROC_'+pageID+'.png')


def WOE(dim_name, data, dim, bucket_num=10, auto=True, bond_num=[]):
    # (trainx, trainy, testx, testy) = ld.getTrainTest(data)
    m = data.shape[0]
    X = data[:, dim]
    y = data[:, -1]
    tot_bad = np.sum(y == 1)
    tot_good = np.sum(y == 0)
    data = np.column_stack((X.reshape(m, 1), y.reshape(m, 1)))
    cnt_bad = []
    cnt_good = []
    min = np.min(data[:, 0])
    max = np.max(data[:, 0])
    if auto == True:

        index = np.linspace(min, max, bucket_num + 1)
    else:
        index = bond_num
        bucket_num = bond_num.shape[0] - 1
    data_bad = data[data[:, 1] == 1, 0]
    data_good = data[data[:, 1] == 0, 0]
    eps = 1e-8
    for i in range(bucket_num):
        if i < bucket_num - 1:
            cnt_bad.append(1.0 * np.sum(np.bitwise_and(data_bad > index[i] - eps, data_bad < index[i + 1])))
            cnt_good.append(1.0 * np.sum(np.bitwise_and(data_good > index[i] - eps, data_good < index[i + 1])))
        else:
            cnt_bad.append(1.0 * np.sum(np.bitwise_and(data_bad > index[i] - eps, data_bad < index[i + 1] + eps)))
            cnt_good.append(1.0 * np.sum(np.bitwise_and(data_good > index[i] - eps, data_good < index[i + 1] + eps)))
    bond = np.array(index)
    cnt_bad = np.array(cnt_bad)
    cnt_good = np.array(cnt_good)

    # cnt_bad[cnt_bad==0] += 1
    # cnt_good[cnt_good==0] += 1


    length = cnt_bad.shape[0]
    for i in range(length):
        j = length - i - 1
        if j != 0:
            if cnt_bad[j] == 0 or cnt_good[j] == 0:
                cnt_bad[j - 1] += cnt_bad[j]
                cnt_good[j - 1] += cnt_good[j]
                cnt_bad = np.append(cnt_bad[:j], cnt_bad[j + 1:])
                cnt_good = np.append(cnt_good[:j], cnt_good[j + 1:])
                bond = np.append(bond[:j], bond[j + 1:])
    if cnt_bad[0] == 0 or cnt_good[0] == 0:
        cnt_bad[1] += cnt_bad[0]
        cnt_good[1] += cnt_good[0]
        cnt_bad = cnt_bad[1:]
        cnt_good = cnt_good[1:]
        bond = np.append(bond[0], bond[2:])
    woe = np.log((cnt_bad / tot_bad) / (cnt_good / tot_good))
    IV = ((cnt_bad / tot_bad) - (cnt_good / tot_good)) * woe
    IV_tot = np.sum(IV)
    bond_str = []
    for b in bond:
        bond_str.append(str(b))
    # print 'IV = %f ' % IV_tot

    cnt_bad_list = cnt_bad.tolist()
    cnt_good_list = cnt_good.tolist()
    cnt_rate = list()
    for index_cnt, cnt in enumerate(cnt_bad):
        cnt_rate.append(float('%.04f'%(cnt/(cnt+cnt_good_list[index_cnt]))))

    out_1 = 'python_out "%s":%s' % (dim_name.strip(), bond_str)
    out_2 = '"cnt_good":%s,"cnt_bad":%s,"cnt_bad_rate":%s' % (cnt_good.tolist(), cnt_bad.tolist(), cnt_rate)
    out_3 = '"woe":%s,"IV":%s' % (woe.tolist(), IV_tot.tolist())
    print (out_1 + ', '+  out_2 +', '+  out_3)

    write_1 = 'python_out "%s":%s' % (dim_name.strip(), bond_str)
    write_2 = '"cnt_good":%s\t"cnt_bad":%s\t"cnt_bad_rate":%s' % (cnt_good.tolist(), cnt_bad.tolist(), cnt_rate)
    write_3 = '"woe":%s\t"IV":%s' % (woe.tolist(), IV_tot.tolist())
    write_data(OutModelFileName, write_1 + '\t'+  write_2 +'\t'+  write_3)

    return IV_tot, IV, woe, bond, cnt_rate, cnt_good, cnt_bad


def score(X_sample, theta, data, bucket_num_list, auto_list, bond_num_list):
    A = 481.86
    B = 28.85
    eps = 1e-8
    if X_sample.shape[0] != theta.shape[0] - 1:
        print ('error: dimension not match')
        return
    base_score = A - B * theta[0]
    old_base = base_score
    for i in range(X_sample.shape[0]):
        xi = X_sample[i]
        (IV_tot, IV, woe, bond, cnt_rate, cnt_good, cnt_bad) = WOE(data, i, bucket_num_list[i], auto_list[i],
                                      bond_num_list[i])
        # print 'bond', bond
        flag = 0
        index = -1
        # print
        for j in range(bond.shape[0] - 1):
            if (j < bond.shape[0] - 2 and xi > bond[j] - eps and xi < bond[j + 1]) or (
                            j == bond.shape[0] - 2 and xi > bond[j] - eps):
                flag = 1
                index = j
                # print 'woe=', woe[j]
        # print
        if flag == 0:
            # print 'error: out of range'
            if xi < bond[0]:
                print ('warning: less than minimum')
                index = 0
        base_score -= (B * theta[i + 1] * woe[index])
        print (B * theta[i + 1] * woe[index])
    print ('base_score = %f ' % old_base)
    print ('score = %f ' % base_score)
    return base_score


def PSI(trainx, testx, trainy, testy, theta):

    testx = np.column_stack((np.ones((testx.shape[0], 1), dtype=float), testx))
    n = theta.shape[0]
    testp = cost.sigmoidFunction(testx, theta.reshape(n, 1)) # prob

    trainx = np.column_stack((np.ones((trainx.shape[0], 1), dtype=float), trainx))
    n = theta.shape[0]
    trainp = cost.sigmoidFunction(trainx, theta.reshape(n, 1)) # prob

    probability_train_list = list()
    cus_num_train_list = list()
    for index_tp, tp in enumerate(trainp):
        probability_train_list.append(tp[0])
        cus_num_train_list.append(index_tp)

    data_left = pd.DataFrame({'probability':probability_train_list, 'cus_num':cus_num_train_list, 'y':trainy},
                              columns=['probability', 'cus_num', 'y'])

    probability_test_list = list()
    cus_num_test_list = list()
    for index_tp, tp in enumerate(testp):
        probability_test_list.append(tp[0])
        cus_num_test_list.append(index_tp)

    data_right = pd.DataFrame({'probability':probability_test_list, 'cus_num':cus_num_test_list, 'y':testy},
                              columns=['probability', 'cus_num', 'y'])

    non_computed = None
    plot_image = True
    check_cols = ['probability', 'y', 'cus_num']
    if non_computed != None and type(non_computed) == str:
        check_cols += [non_computed]
        data_left = data_left[~data_left[non_computed] == True].copy()
        data_right = data_right[~data_right[non_computed] == True].copy()
    elif non_computed == None:
        data_left = data_left.copy()
        data_right = data_right.copy()
    else:
        raise ValueError('non_computed must be a str.')
    for col in check_cols:
        if col not in data_left.columns or col not in data_right.columns:
            raise ValueError('There is no column %s of data' % col)

    """Drop NaN and sort values by column 'probability'."""
    data_left = data_left.loc[data_left['probability'].notnull(), check_cols]
    data_right = data_right.loc[data_right['probability'].notnull(), check_cols]
    data_left.sort_values(by=['probability'], inplace=True, ascending=False)
    data_right.sort_values(by=['probability'], inplace=True, ascending=False)
    data_left.reset_index(drop=True, inplace=True)
    data_right.reset_index(drop=True, inplace=True)

    """Discrete probability value."""
    break_points_left = [int(data_left.shape[0] * i) for i in np.arange(0.0, 1.1, 0.1)]
    break_points_right = [int(data_right.shape[0] * i) for i in np.arange(0.0, 1.1, 0.1)]
    quantile_str = [str(i) + '%' for i in np.arange(10, 110, 10)]
    for i in range(len(break_points_left)):
        if i == len(break_points_left) - 1:
            break
        else:
            data_left.loc[break_points_left[i]:break_points_left[i + 1], 'quantile'] = quantile_str[i]
            data_right.loc[break_points_right[i]:break_points_right[i + 1], 'quantile'] = quantile_str[i]

    """Count psi of bad & good sample."""
    count_left = data_left.groupby(['quantile', 'y']).count()['cus_num'].unstack().fillna(value=0.0)
    count_right = data_right.groupby(['quantile', 'y']).count()['cus_num'].unstack().fillna(value=0.0)


    a = 0

    try:
        if count_left[1].sum() == 0 or count_right[1].sum() == 0:
            a = 1
            return 0, 1
        if count_left[0].sum() == 0 or count_right[0].sum() == 0:
            a = 1
            return 0, 1

    except:
        a = 1
        return 0, 1

    count_left['bad_ratio'] = count_left[1] / count_left[1].sum()
    count_right['bad_ratio'] = count_right[1] / count_right[1].sum()
    count_left['good_ratio'] = count_left[0] / count_left[0].sum()
    count_right['good_ratio'] = count_right[0] / count_right[0].sum()
    count_final = pd.merge(count_left, count_right, left_index=True,
                           right_index=True, suffixes=['_left', '_right'])
    count_final['psi_bad'] = (count_left['bad_ratio'] - count_right['bad_ratio']) * np.log(count_left['bad_ratio'] / count_right['bad_ratio'])
    count_final['psi_good'] = (count_left['good_ratio'] - count_right['good_ratio']) * np.log(count_left['good_ratio'] / count_right['good_ratio'])
    count_final = count_final.reindex(quantile_str)
    average_psi = (count_final['psi_bad'].replace([np.inf, np.nan], 0.0).sum() + count_final['psi_good'].replace([np.inf, np.nan], 0.0).sum()) / 2

    print ('Average PSI:%f' % average_psi )
    write_data(OutModelFileName, 'Average PSI:%f' % average_psi)


    if 0 in list(count_right['bad_ratio'].values) or 0 in list(count_right['good_ratio']):
        a = 1


    return average_psi, a

def get_percent(score_num, lower_bound, higher_bound, total_num):
    gt = [i >= lower_bound[0] for i in score_num]
    lt = [i <= higher_bound[0] for i in score_num]
    len_t = 0
    for index_t, t in enumerate(gt):
        if t[0] and lt[index_t][0]:
            len_t += 1
    return len_t/total_num


def calcVIF(data, columns_dim):
    m = data.shape[0]
    n = data.shape[1] - 1

    VIF_list = list()

    if n < 2:
        print ('python_out \"VIF\":-1')
        VIF_list.append(-1)
        return
    for i in range(n):
        newy = data[:, i].reshape(m, 1)
        newX = np.column_stack((data[:, :i], data[:, i + 1:n]))
        # print data[:, 0]

        theta = train.startTrain(newX, newy, 0, type='linear').reshape(n, 1)
        # print theta
        newX = np.column_stack((np.ones((m, 1), dtype=float), newX))
        predict = newX.dot(theta).reshape(m, )
        newy = newy.reshape(m, )
        # VIF.append(np.sum(np.square(newy-np.mean(newy)))/np.sum(np.square(predict-newy)))
        VIF = np.sum(np.square(newy - np.mean(newy))) / np.sum(np.square(predict - newy))

        VIF_list.append(':%f' % VIF)

        print ('python_out \"VIF\":%f' % VIF + " " + str(columns_dim[i]))
        # return VIF
    # print('python_out' + str(VIF_list))

    write_data(OutModelFileName, 'python_out' + str(VIF_list))

    return


if __name__ == '__main__':
    pass



# coding:utf-8


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt



from lapras.DealConfig import checkDataFileName, makeModelDataFillnaFileName, scoreCardDataFileName, comParamCorrFileName
from lapras.DealConfig import A, B, theta, woe_all, cut_points_step, assist_num
from lapras.ReportConfig import  cut_points_step, assist_num, score_bond

from lapras.ReportCode.ScoreParam import get_score_csv, get_score_max_min, get_score_bond
from lapras.ReportCode.MakeModelDoc import make_model_doc,count_point, plt_show



'''
模型效果展示  

整合数据 得到对应图表， 并打印部分统计数据
'''

def show_pic(score_bond):
    # 保存入模变量 变量相似性 得分结果

    # 计算 区间数量 区间坏账率
    x, ticks, y_count, y_rate = count_point(checkDataFileName, score_bond)

    # 画图显示 区间数量 区间坏账率
    plt_show(x, ticks, y_count, y_rate)

def show_model_pic(score_bond):


    get_score_csv(makeModelDataFillnaFileName, scoreCardDataFileName, comParamCorrFileName, checkDataFileName)
    if not score_bond:
        # 得到计分区间
        base_score_min, base_score_max = get_score_max_min(A, B, theta, woe_all)
        # 自动划分区间
        score_bond = get_score_bond(cut_points_step, assist_num=assist_num, minpoint=base_score_min, maxpoint=base_score_max, checkDataFileName=checkDataFileName)

    print(score_bond)

    # 展示效果
    show_pic(score_bond)
    # 为写模型部署文档准备表格数据
    # make_model_doc(params, theta, woe_all, param_bond, B)


if __name__ == '__main__':
    # get_score_csv(makeModelDataFillnaFileName, scoreCardDataFileName, comParamCorrFileName, checkDataFileName)
    # if not score_bond:
    #     # 得到计分区间
    #     base_score_min, base_score_max = get_score_max_min(A, B, theta, woe_all)
    #     # 自动划分区间
    #     score_bond = get_score_bond(cut_points_step, assist_num=assist_num, minpoint=base_score_min, maxpoint=base_score_max, checkDataFileName=checkDataFileName)
    # print(score_bond)
    # # 展示效果
    # show_pic(score_bond)
    # # 为写模型部署文档准备表格数据
    # # make_model_doc(params, theta, woe_all, param_bond, B)

    show_model_pic()


import os
import random as rand
import sys
import numpy as np
import pandas as pd
from lapras.DealConfig import makeModelDataFillnaFileName, OnlyModelDataFillnaFileName
from lapras.ModelCode.DealParamConfig import params
from lapras.Settings import col_head


def get_only_model_data():
    param_list = col_head + params
    df_data = pd.read_csv(makeModelDataFillnaFileName)
    df_data = df_data.dropna(how='all')
    df_data[param_list].to_csv(OnlyModelDataFillnaFileName, index=None, encoding='utf-8')


def load_data(path):
    companylist = list()
    f=open(path)

    data=[]
    columns = list()
    for index_l, line in enumerate(f):
        if index_l==0:
            columns = line
            continue

        cols=line.strip().split(',')
        companylist.append(cols[0])

        line_data=[float(x) for x in cols[2:]]

        line_data.append(float(cols[1]))

        data.append(line_data)
    f.close
    return np.array(data), columns.split(','), companylist





def rankData(data, dim, bond, woe):
    data_dim = data[:, dim]
    m = data_dim.shape[0]
    b = bond.shape[0]
    eps = 1e-8
    for i in range(m):
        xi = float(data_dim[i])
        bond_index = -1
        for j in range(b - 1):
            if (j < b - 2 and xi > bond[j] - eps and xi < bond[j + 1]) or (
                                j == b - 2 and xi > bond[j] - eps and xi < bond[j + 1] + eps):
                bond_index = j
        data_dim[i] = woe[bond_index]
    return data


def getTrainTest(data, sep=-1):
    if sep == -1:
        np.random.shuffle(data)
        sep = int(data.shape[0] * 0.7)
    trainx = data[:sep, :-1]
    trainy = data[:sep, -1]
    testx = data[sep:, :-1]
    testy = data[sep:, -1]
    return 1. * trainx, 1. * trainy, 1. * testx, 1. * testy


if __name__ == '__main__':
    # a = ['1', '2', '3']
    # print str(a[:-1])[1:-1].replace('\'', '')
    # print str([0 for i in range(3)])[1:-1]
    # a = np.array([[1, 2], [2, 3]])
    # print a.tolist()
    # x = 27
    # y = 100.000001
    # print x < y
    print (-np.inf - 3)



# coding:utf-8

from lapras.ModelConfig import bin_params_dict
from lapras.ModelConfig import  minLeaf, treeDepth, bucketnum

params = list(bin_params_dict.keys())
bin_params =  [bin_params_dict.get(para) for para in params]


# coding:utf-8


def write_data(filename, data):
    if isinstance(data, str):
        with open(filename, 'a') as f:
            f.write(data + '\n')
        f.close()
    else:
        pass


def deal_data(filename):
    try:
        with open(filename, 'r+') as f:
            f.truncate()
        f.close()
    except:
        pass





# coding:utf-8

import numpy as np
import scipy.optimize as opt
from lapras.ModelCode.utils import cost


def startTrain(trainx, trainy, lamb, type='logistic'):
    m = trainx.shape[0]
    n = trainx.shape[1] + 1
    # print m,n
    X = np.column_stack((np.ones((m, 1), dtype=float), trainx))
    # print X
    y = trainy
    min_J = np.inf
    min_theta = np.zeros(n)
    eps = 0.1
    if type == 'logistic':
        f = cost.f_logistic
        fgrad = cost.fgrad_logistic
    elif type == 'linear':
        f = cost.f_linear
        fgrad = cost.fgrad_linear
    for i in range(30):
        init_theta = np.random.rand(n) * 2.0 * eps - eps
        # print init_theta
        theta = opt.fmin_cg(f, init_theta, fprime=fgrad, args=(X, y, lamb), disp=False)
        # print theta
        J = f(theta, X, y, lamb)
        if J < min_J:
            min_J = J
            min_theta = theta
    return min_theta


if __name__ == '__main__':
    # x = np.ones((3, 1), dtype=int)
    # print np.asarray(x)


    # def f(x):
    #     a = x[0]
    #     b = x[1]
    #     return (a + b) ** 2 + 10


    # def fGrad(x):
    #     a = x[0]
    #     b = x[1]
    #     return np.mat('%s;%s' % (2 * (a + b), 2 * (a + b)))


    # init_x = np.mat('3;3').T
    # print init_x
    # print fGrad(init_x).T
    # res = opt.fmin_cg(f, init_x, fprime=fGrad)
    # print res

    trainx = np.array([[1], [2], [3], [4], [5], [6]])
    m = trainx.shape[0]
    trainy = np.array([1, 1, 1, 0, 0, 0])
    theta = startTrain(trainx, trainy, 0)
    print (cost.sigmoidFunction(np.column_stack((np.ones((m, 1), dtype=float), trainx)), theta))
    print (theta)



# coding:utf-8

# 设置环境变量
import pandas as pd

from lapras.ModelCode import AnalyzeData

from lapras.DealConfig import makeModelDataFillnaFileName, ruleFilePriName
from lapras.DealConfig import OnlyModelDataFillnaFileName, ruleModelFilePriName, ruleModelFileName
from lapras.ModelCode.loadData import get_only_model_data


def run_model(datafilepath,run_type_cus=-1):
    # 建模
    if run_type_cus == 0:
        AnalyzeData.state_data(datafilepath, ruleFilePriName, 0)
    elif run_type_cus == 1:
        get_only_model_data()
        AnalyzeData.state_data(OnlyModelDataFillnaFileName, ruleModelFileName, 1)
    elif run_type_cus == 2:
        get_only_model_data()
        while True:
            perf, ks, ks_train, ROC,ks_max_index, average_psi, a, min_theta = AnalyzeData.state_data(OnlyModelDataFillnaFileName, ruleModelFilePriName, 1)
            ks_max = ks[ks_max_index][0]
            ks_train_max = ks_train[ks_max_index][0]
            if ks_max < 0.25:
                continue
            if ks_max - ks_train_max > 0.03:
                continue
            if average_psi < 0.1 and min_theta > 0 and a == 0:
                perf.KS_draw(ks[ks_max_index][1], ks[ks_max_index][2])
                perf.ROC_draw(ROC[ks_max_index][1], ROC[ks_max_index][2])
                break


if __name__ == '__main__':

    # 规则 建模模块
    run_model(run_type_cus=-1)
    # 0是所有变量的单次跑
    # 1是单次建模
    # 2是循环直到合适的数据



# coding:utf-8



import pandas as pd
import matplotlib.pyplot as plt
import matplotlib
from collections import Counter
from lapras.Settings import col_head
from lapras.DealConfig import params, theta, woe_all, param_bond, B, score_bond, checkDataFileName
from lapras.ReportConfig import x_label, y_label_left, y_label_right, graph_title, line_flag

'''
制作 模型部署文档 的准备部分，准备变量和对应模型分

需要  config 或 settings 中 含有 params, theta, woe_all, param_bond
需要  config  或 setting 中 含有 B, score_bond check_data_storeout（文件）
需要  文件check_data_storeout 中包含 y 和 score 即样本表现和样本评分
'''

def get_max_int(int_data):
    return int(str(int(str(int_data)[0])+1) + '0' * (len(str(int_data))-1))

def make_model_doc(params, theta, woe_all, param_bond, B):
    '''
    打印模型部署文档 表格中的区间数据及最值差
    '''
    params_print = list()
    param_score_all_list = list()
    param_bond_all_list = list()
    max_min_list = list()
    for index_woe, woe in enumerate(woe_all):
        # print (params_list[2:][index_woe])
        param_score_list = list()
        param_bond_list = param_bond.get(params[index_woe])

        for index_woe_one, woe_one in enumerate(woe):
            params_print.append(params[index_woe])
            param_score = -woe_one * theta[1:][index_woe] * B
            param_score_list.append(param_score)
            param_score_all_list.append(param_score)
            len_param_bond_list = len(param_bond_list)
            if index_woe_one == 0:
                bond_one = "<%s"%(param_bond_list[index_woe_one])
            elif index_woe_one == len_param_bond_list:
                bond_one = ">=%s" % (param_bond_list[index_woe_one-1])
            else:
                bond_one = "[%s, %s)" % (param_bond_list[index_woe_one-1],param_bond_list[index_woe_one])
            param_bond_all_list.append(bond_one)
            # print(param_score)
        max_min = max(param_score_list) - min(param_score_list)
        max_min_list.append(max_min)

    print( '模型分')
    for index_param, param_score in enumerate(param_score_all_list):
        print('%s\t%s\t%s\t%s'%(params_print[index_param], '中文名',param_bond_all_list[index_param], param_score))

    print('最大分差')
    for index_param, max_min in enumerate(max_min_list):
        print('%s\t%s'%(params[index_param], max_min))

def count_point(check_data_file, score_bond):
    '''
    读取分数文件 包含 y base_score 字段
    y 为表现 base_score 为 评分
    :return: 各区间好坏样本数量
    '''
    scoredata = pd.read_csv(check_data_file)
    labels = list(range(len(score_bond) - 1))
    baddf = scoredata[scoredata[col_head[1]] == 1]
    gooddf = scoredata[scoredata[col_head[1]] == 0]

    # 获得 好坏样本 总样本
    badstat = pd.cut(baddf['base_score'], bins=score_bond, labels=labels, include_lowest=True)
    goodstat = pd.cut(gooddf['base_score'], bins=score_bond, labels=labels, include_lowest=True)
    allstat = pd.cut(scoredata['base_score'], bins=score_bond, labels=labels, include_lowest=True)

    # 统计各分数段样本数量
    badstat_result = pd.value_counts(badstat, sort=False)
    goodstat_result = pd.value_counts(goodstat, sort=False)
    allstat_result = pd.value_counts(allstat, sort=False)

    bad_count = badstat_result.tolist()
    good_count = goodstat_result.tolist()
    y_count = allstat_result.tolist()
    print('bad: %s' % bad_count)
    print('good: %s' % good_count)
    print('all: %s' % y_count)
    all_rate = ['%.2f%%'%(y_c/sum(y_count) *100) for y_c in y_count]
    print('all_rate: %s' % all_rate)


    # 计算区间坏账率
    ticks =  ['(%d,%d]' % (score_bond[i], score_bond[i+1]) for (i,x) in enumerate(score_bond) if i<len(score_bond)-1]
    score_stat_df = pd.DataFrame({'labels':labels,'bad_count':bad_count,'good_count':good_count,'y_count':y_count})
    # score_stat_df['ratio'] = score_stat_df['bad_count']/(score_stat_df['good_count']+score_stat_df['bad_count'])
    score_stat_df['y_rate'] = score_stat_df['bad_count']/(score_stat_df['y_count'])
    # score_stat_df['ratio'] = score_stat_df['bad_count']/len( score_stat_df['bad_count'])
    x = score_stat_df['labels']
    y_rate = score_stat_df['y_rate']
    y_rate_cent = ['%.2f%%'%(y_i *100) for y_i in list(y_rate)]
    print ('bad_rate: %s'%(y_rate_cent))
    base_score_count = Counter(list(scoredata['base_score']))
    print ('重复分重复次数排序（前15个） %s'%base_score_count.most_common()[:15])
    # x: 区间分段 1,2,3,4
    # ticks: 区间名称['[300, 400)', '[400, 500)',  '[500, 1000)']
    # y_count: 区间 数量， 表示评分在此区间内的样本数量
    # y_rate: 区间 坏账率
    return x, ticks, y_count, y_rate

def plt_show(x, ticks,y_count, y_rate, title=None):
    '''
    画 柱状图 和 折线图
    :param x: 区间分段 1,2,3,4
    :param ticks: 区间名称['[300, 400)', '[400, 500)',  '[500, 1000)']
    :param y_count: 区间 数量， 表示评分在此区间内的样本数量
    :param y_rate: 区间 坏账率
    '''
    # 设置字体、图形样式
    # sns.set_style("whitegrid")
    matplotlib.rcParams['font.sans-serif'] = ['SimHei']
    matplotlib.rcParams['font.family'] = 'sans-serif'
    matplotlib.rcParams['axes.unicode_minus'] = False
    matplotlib.fontsize='15'

    y1 = y_count
    y2 = y_rate
    # 设置图形大小
    plt.rcParams['figure.figsize'] = (18.0, 9.0)

    fig = plt.figure()

    # 画柱子
    ax1 = fig.add_subplot(111)
    # alpha透明度， edgecolor边框颜色，color柱子颜色 linewidth width 配合去掉柱子间距
    ax1.bar(x, y1, alpha=0.8, edgecolor='k', color='#836FFF',linewidth=1, width =1)
    # 获取 y 最大值 最高位 + 1 的数值 比如 201取300，320取400，1800取2000
    y1_lim = get_max_int(max(y1))
    # 设置 y轴 边界
    ax1.set_ylim([0, y1_lim])
    # 设置 y轴 标题
    ax1.set_ylabel(y_label_left, fontsize='15')
    ax1.set_xlabel(x_label,fontsize='15')
    # 将分值标注在图形上
    for x_i, y_i in  zip(x, y1):
        ax1.text(x_i, y_i + y1_lim/20, str(y_i), ha='center', va='top', fontsize=13, rotation=0)

    # 设置标题
    if title != None:
        ax1.set_title(title, fontsize='20')
    else:
        ax1.set_title(graph_title, fontsize='20')
    plt.yticks(fontsize=15)
    # plt.xticks(x, y)
    plt.xticks(fontsize=12)

    # 画折线图
    if line_flag:
        ax2 = ax1.twinx()  # 这个很重要噢
        ax2.plot(x, y2, 'r', marker='*', ms=0)

        # ax2.set_xlim([-0.5, 3.5])
        try:
            y2_lim = (int(max(y2) * 10) + 1) / 10
        except:
            y2_lim = 1
        ax2.set_ylim([0, y2_lim])
        ax2.set_ylabel(y_label_right, fontsize='15')
        ax2.set_xlabel(x_label,fontsize='15')
        for x_i, y_i in  zip(x, y2):
            ax2.text(x_i, y_i+y2_lim/20 , '%.2f%%'%(y_i *100), ha='center', va='top', fontsize=13, rotation=0)
    plt.yticks(fontsize=15)
    plt.xticks(x, ticks)
    plt.xticks(fontsize=15)

    # 是否显示网格
    plt.grid(True)

    # 保存图片 dpi为图像分辨率
    # plt.savefig('分数分布及区间坏账率.png', dpi=600, bbox_inches='tight')
    # 显示图片
    plt.show()



if __name__ == '__main__':
    # 为写模型部署文档准备表格数据
    make_model_doc(params, theta, woe_all, param_bond, B)
    # make_model_doc(params, theta, woe_all, param_bond, B)

    # 计算 区间数量 区间坏账率
    x, ticks, y_count, y_rate = count_point(checkDataFileName, score_bond)

    # 画图显示 区间数量 区间坏账率
    plt_show(x, ticks,y_count, y_rate)


# coding:utf-8

import numpy as np
from sklearn import tree
from sklearn.tree import DecisionTreeClassifier
import re

def decisionTreeSplit(datax, datay, DepthOfTree, MinSamplesLeaf):
    groupdt = DecisionTreeClassifier(criterion='entropy', splitter='best', max_depth=DepthOfTree,
                                     min_samples_leaf=MinSamplesLeaf)
    groupdt.fit(datax.reshape((datax.shape[0], 1)), datay.reshape((datay.shape[0], 1)))
    dot_data = tree.export_graphviz(groupdt, out_file=None, )
    pattern = re.compile('<= (.*?)\\\\nentropy', re.S)
    split_num = re.findall(pattern, dot_data)
    a = [float(b) for b in split_num]
    return sorted(a)

if __name__ == "__main__":
    datax=np.array([int(line.strip()) for line in open('datax.csv')])
    datay=np.array([int(line.strip()) for line in open('datay.csv')])
    aaa = decisionTreeSplit(datax,datay,10,100)
    print (aaa)




# coding:utf-8

import matplotlib.pyplot as plot
from matplotlib.ticker import  MultipleLocator

def woe_draw(bond, sum_bad_percent, count_good):
    ks_bucket_x = [i + 1.5 for i in range(len(bond) - 1)]
    bond = list(bond)
    bond.insert(0, '0')
    ks_bucket_num = [i / 2 for i in range(2, int(ks_bucket_x[-1] * 2) + 2)]

    fig = plot.figure()
    ax1 = fig.add_subplot(111)
    ax1.plot(ks_bucket_x, sum_bad_percent, marker="*", linewidth=3, color="orange")

    for i, (_x, _y) in enumerate(zip(ks_bucket_x, sum_bad_percent)):
        plot.text(_x, _y, sum_bad_percent[i], color='black', fontsize=10, )

    ax1.set_ylim(bottom=-5, top=15)
    ax1.set_title('Trend Of Feature')

    ax1.set_xticks(ks_bucket_num)
    ax1.set_xticklabels(bond, fontsize=12)
    xmajorLocator = MultipleLocator(1)
    ax1.xaxis.set_major_locator(xmajorLocator)

    ax1.legend(["woe"], loc="upper right")
    ax1.grid(True)
    ax2 = ax1.twinx()
    plot.bar(ks_bucket_x, count_good, alpha=0.3, color='blue', label='count_good')
    ax2.set_ylim(bottom=0, top=1000000)  # [9.0, 1.0, 875440.0]
    ax2.legend(loc=2)
    plot.show()



# coding:utf-8

import os
import time
from lapras.Settings import makeModelDataFillnaFile, ModelDataFillnaFile, ruleModelFilePri, ruleModelFile, OutModelFile
from lapras.Settings import csvFile, ruleFilePri, scoreCardDataFile, comParamCorrFile, checkDataFile,scoreCardParamFile,scoreCardPklFile,woeFile
from lapras.ReportConfig import cut_points_step, assist_num, score_bond
from lapras.ModelConfig import A, B, eps
from lapras.ReportConfig import ParamsShow, ModelName
from lapras.ReportCode.ReportConfigUtil import get_param_bond_woe_all_theta, get_GBK_Name

# 文件部分

if not os.path.exists(csvFile):
    os.makedirs(csvFile)
# 全部变量的 csv数据文件
makeModelDataFillnaFileName = csvFile + makeModelDataFillnaFile + '.csv'
# 只有入模变量的 csv 文件
OnlyModelDataFillnaFileName = csvFile + ModelDataFillnaFile + '.csv'


# 分箱 好坏区间及占比 IV
# 全量
ruleFilePriName =  csvFile + ruleFilePri + '.csv'
# 只看入模变量的
# 决策树分
ruleModelFilePriName =  csvFile + ruleModelFilePri + '.csv'
# 手动分
ruleModelFileName =  csvFile + ruleModelFile + '.csv'


OutModelFileName =  csvFile + OutModelFile + time.strftime("%Y%m%d_%H%M%S", time.localtime())


#  当前入模变量的 csv数据文件
scoreCardDataFileName =  csvFile + scoreCardDataFile + 'file' + '.csv'
# 模型变量相关性
comParamCorrFileName =  csvFile + comParamCorrFile + 'file' + '.csv'
# 模型变量评分
checkDataFileName =  csvFile + checkDataFile + 'file' + '.csv'
# 评分卡参数文件
scoreCardParamFileName =  csvFile + scoreCardParamFile + 'file' + '.txt'
# 评分卡pkl文件
scoreCardPklFileName =  csvFile + scoreCardPklFile + 'file' + '.pkl'
# woe明细文件路径
woeFileName =  csvFile + woeFile + 'file' + '.csv'



# 生成报告需要的模型文件
# 单变量分析的model文件， 主要用于模型单变量查看
singleModelName = csvFile + ModelName

try:
    with open(singleModelName, 'r') as f:
        params_data = f.read()
    f.close()
    params, param_bond, woe_all, theta, bond_list = get_param_bond_woe_all_theta(params_data)
except:
    params, param_bond, woe_all, theta, bond_list = (0, 0, 0, 0, 0)





# coding:utf-8

import numpy as np


def sigmoidFunction(X, theta):
    return 1.0 / (1.0 + np.exp(-X.dot(theta)))


def f_logistic(theta, X, y, lamb):
    (m, n) = X.shape
    # print theta
    h = sigmoidFunction(X, theta.reshape(n, 1)).reshape(m, )
    # print h
    J = -(1.0 / m) * (np.inner(y, np.log(h)) + np.inner((1.0 - y), np.log(1.0 - h))) + (lamb / 2.0 / m) * np.sum(
        np.square(theta[1:]))
    # print J
    return J


def fgrad_logistic(theta, X, y, lamb):
    (m, n) = X.shape
    h = sigmoidFunction(X, theta.reshape(n, 1)).reshape(m, )
    temp = lamb * theta
    temp[0] = 0.0
    grad = (1. / m) * (X.T.dot((h - y).reshape(m, 1)) + temp.reshape(n, 1))
    return grad.reshape(n, )


def f_linear(theta, X, y, lamb):
    (m, n) = X.shape
    J = (0.5 / m) * np.sum(np.square(y - X.dot(theta.reshape(n, 1))))
    return J


def fgrad_linear(theta, X, y, lamb):
    (m, n) = X.shape
    grad = (1.0 / m) * ((X.dot(theta.reshape(n, 1)) - y) * X)
    return np.sum(grad, 0)


if __name__ == '__main__':
    # import scipy.optimize as opt
    #
    #
    # def f(x, y, z):
    #     return (x + y) ** 2 + z
    #
    #
    # def fGrad(x, y, z):
    #     return 2 * (x + y)
    #
    #
    # res = opt.fmin_cg(f, 10, fprime=fGrad, args=(5, 3))
    # print res
    # x = np.mat('3;3;3')
    # y = np.column_stack((np.mat('2;2;2'), x))
    # print y
    # print np.sum(x)
    # print np.sum(np.power(x, 2))
    # x = np.mat('1,2,3')
    # print x.tolist()[0]
    # a = np.array([[1, 2, 3], [2, 3, 4]])
    # b=np.array([2,4])
    # b=b.reshape(2,1)
    # print a
    # print b
    # print a*b
    # print np.sum(a, 0)
    a = np.mat('1,2,3;2,3,4')
    b = np.mat('2;4')
    print (np.array(a) * np.array(b))



# coding:utf-8

import math
import os

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import roc_auc_score
from sklearn.metrics import roc_curve

from lapras.DealConfig import checkDataFileName
from lapras.Settings import col_head




def ks_calc_cross(pred, y_label):
    '''
    功能: 计算KS值，输出对应分割点和累计分布函数曲线图
    输入值:
    data: 二维数组或dataframe，包括模型得分和真实的标签
    pred: 一维数组或series，代表模型得分（一般为预测正类的概率）
    y_label: 一维数组或series，代表真实的标签（{0,1}或{-1,1}）
    输出值:
    'ks': KS值，'crossdens': 好坏客户累积概率分布以及其差值gap
    '''
    crossfreq = pd.crosstab(pred, y_label)
    crossdens = crossfreq.cumsum(axis=0) / crossfreq.sum()
    crossdens['gap'] = abs(crossdens[0] - crossdens[1])
    ks = crossdens[crossdens['gap'] == crossdens['gap'].max()]
    return float(ks['gap'])


def ks_calc_auc(pred, y_label):
    '''
    功能: 计算KS值，输出对应分割点和累计分布函数曲线图
    输入值:
    data: 二维数组或dataframe，包括模型得分和真实的标签
    pred: 一维数组或series，代表模型得分（一般为预测正类的概率）
    y_label: 一维数组或series，代表真实的标签（{0,1}或{-1,1}）
    输出值:
    'ks': KS值
    '''
    fpr, tpr, thresholds = roc_curve(y_label, pred)
    ks = max(tpr - fpr)
    return ks


'''
运行代码主文件

使用时请注释掉不需要运行的行
'''

def performance(checkDataFileName=checkDataFileName,bad_label=col_head[1]):

    df = pd.read_csv(checkDataFileName, engine='python', encoding='utf-8')
    df["rank"] = 1000 - df['base_score']

    # AUC值 KS值
    precisions, recalls, thresholds = precision_recall_curve(df[bad_label], df["rank"])
    fpr, tpr, thresholds = roc_curve(df[bad_label], df["rank"])
    auc = roc_auc_score(df[bad_label], df["rank"])
    ks = max(tpr - fpr)
    print("KS: %.4f" % (ks))
    print("AUC: %.4f"  % (auc))

    # 设置图形大小
    plt.rcParams['figure.figsize'] = (12.0, 8.0)
    plt.rcParams.update({'font.size': 20})

    # 绘制AUC曲线
    # plt.legend(('train_set', 'test_set', 'OOT1', 'OOT2', 'OOT3'), loc='lower right')
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('ROC Line')
    plt.plot(fpr[:-1], tpr[:-1])
    plt.show()

    # 绘制KS曲线
    x_axis = [i / len(tpr[:-1]) for i in range(len(tpr[:-1]))]
    position = np.argmax(tpr - fpr)
    plot_position = position / len(tpr[:-1])
    plt.plot(x_axis, tpr[:-1])
    plt.plot(x_axis, fpr[:-1])
    plt.plot([plot_position, plot_position], [tpr[position], fpr[position]])
    plt.legend(('True Positive', 'False Positive', 'Max Ks Gap'), loc='upper left')
    plt.xlabel('Percent')
    plt.ylabel('Rate')
    plt.text(0.5, 0.5, 'ks=' + str(round(ks, 3)))
    plt.title('KS Line')
    plt.show()

    # 绘制PR曲线
    plt.plot(recalls[:-1], precisions[:-1])
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.title('PR Line')
    plt.show()

    # 计算PR
    tmp = pd.DataFrame([recalls[:-1], precisions[:-1]]).T
    tmp.columns = ['recalls', 'precisions']
    recalls = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    for recall in recalls:
        presion = 0.002
        d = tmp[(tmp['recalls'] >= recall - presion) & (tmp['recalls'] <= recall + presion)].reset_index(drop=True)
        print(d.iloc[:1, :])


# coding:utf-8

from lapras.ReportCode.SingleParamAnalyze import show_mutil, show_single
from lapras.ModelCode.main import run_model
from lapras.DealConfig import singleModelName, ParamsShow
from lapras.ReportConfig import score_bond
from lapras.ReportCode.ShowModelResult import show_model_pic
from lapras.model_performance import performance


'''
运行代码主文件

使用时请注释掉不需要运行的行
'''

if __name__ == '__main__':

    '''
    建模 
    '''
    '''     决策树分箱 '''
    run_model("model_data.csv",run_type_cus=0)
    '''     手动分箱 确认分箱 '''
    # run_model(run_type_cus=1)
    '''     确认分箱 循环跑 寻找较优解 '''
    # run_model(run_type_cus=2)


    ''' 
    报告分析 
    '''
    # show_single(singleModelName, ParamsShow)  #单变量批量查看

    # show_mutil(singleModelName)   #单变量全部查看


    ''' 
    整体建模报告查看 
    '''
    # show_model_pic(score_bond)

    ''' 
    模型效果查看 
    '''
    # performance()




#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from setuptools import setup, find_packages


setup(
    name='lapras',
    version='0.0.6',
    packages = find_packages(),

    description='scorecard',
    # long_description=open("README.md").read(),
    # long_description_content_type = 'text/markdown',
    url='https://github.com/yhangang',
    author='Badtom',
    author_email='yhangang@gmail.com',
    license='MIT',
    python_requires = '>=3.5',
    install_requires = [
        'numpy >= 1.15.4',
        'pandas',
        'scipy',
        'scikit-learn',
        'seaborn >= 0.10.0',
    ],
    classifiers=[
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    package_data = {
        # If any package contains *.txt or *.rst files, include them:
        # '': ['*.txt', '*.rst', '*.csv'],
        # include any *.msg files found in the 'test' package, too:
        # 'data': ['*.csv'],
    },
    # data_files=[('StartProject.py'),('DealConfig.py'),('model_performance.py'),('ModelConfig.py'),('README.md'),('ReportConfig.py'),('Settings.py')],

    keywords=['scorecard'],
    entry_points={

    },
)


# coding:utf-8

import numpy as np
import pandas as pd
from copy import deepcopy

from lapras.ModelCode.utils import decisionTree as dt
from lapras.ModelCode.utils import perform as perf
from lapras.ModelCode.utils import train
from lapras.ModelCode import loadData as ld
from lapras.ModelCode.DealParamConfig import bin_params, minLeaf, treeDepth, bucketnum
from lapras.ModelCode.utils.FileUtil import write_data, deal_data
from lapras.DealConfig import OutModelFileName


def load_data(path):
    companylist = list()
    with open(path) as f:
        data=[]
        columns = list()
        for index_l, line in enumerate(f):
            if index_l==0:
                columns = line
                continue

            cols=line.strip().split(',')
            companylist.append(cols[0])
            line_data=[float(x) for x in cols[2:]]


            line_data.append(float(cols[1]))

            data.append(line_data)
    f.close()
    return np.array(data), columns.split(','), companylist



def state_data(new_file_name, rule_file_name=None, makemodel=1):
    print ('读取 new_file_name')

    deal_data(OutModelFileName)

    data, columns, companylist = load_data(new_file_name)

    dims = len(columns) -2
    test_case = 1
    init_flag_list = [1]*dims # 特征数 list
    # 分箱方法
    eql_dis_flag_list = [0]*dims # 等段分箱

    if makemodel == 1:
        tree_flag_list = [0] * dims  # 决策树分箱 1采用 0 不采用
        bond_list = deepcopy(bin_params)

    else:
        tree_flag_list = [1] * dims  # 决策树分箱 1采用 0 不采用
        bond_list = [0] * dims

    # 离散桶
    bucketnum_list = [bucketnum]*dims
    # 树深度
    treeDepth_list = [treeDepth]*dims
    # 最小子数目
    # minLeaf_list = [1]*dims
    minLeaf_list = [minLeaf]*dims

    X_dim = len(init_flag_list)

    res_arr = None

    columns_dim = columns[2:]
    woe_list = []
    print('计算 WOE')
    for i in range(X_dim):
        if tree_flag_list[i] == 1:
            bond_list[i] = dt.decisionTreeSplit(data[:, i], data[:, -1], treeDepth_list[i], minLeaf_list[i])
        bond_list[i].insert(0, -np.inf)
        bond_list[i].append(np.inf)
        bond_list[i] = np.sort(np.array(bond_list[i]))
        IV_tot, IV, woe, bond, cnt_rate, cnt_good, cnt_bad = perf.WOE(columns_dim[i], data, i,
                                                                      bucket_num=bucketnum_list[i],
                                                                      auto=(eql_dis_flag_list[i] == 1),
                                                                      bond_num=bond_list[i])
        bond_list[i] = bond
        woe_list.append(woe)

        trend_list = list(cnt_rate)
        trend = ''
        if len(trend_list) > 1:
            for index_cnt_rate, cnt_rate_i in enumerate(trend_list[1:]):
                if cnt_rate_i > trend_list[index_cnt_rate]:
                    trend = trend + '+'
                elif cnt_rate_i < trend_list[index_cnt_rate]:
                    trend = trend + '-'
                else:
                    trend = trend + '='
        else:
            trend = '/'

        dim_arr = np.array([columns_dim[i].strip(), str(list(bond)), str(list(cnt_good)), str(list(cnt_bad)), str(list(cnt_rate)), trend, str(list(woe)), str(list(IV)), float('%.08f' % IV_tot), float('%.03f' % IV_tot)])
        if res_arr is None:
            res_arr = dim_arr
        else:
            res_arr = np.vstack((res_arr, dim_arr))

    columns = ['dim_name', 'bin', 'cnt_good', 'cnt_bad', 'cnt_bad_rate', 'trend','woe', 'IV_i', 'IV', 'IV_sort']
    df = pd.DataFrame(res_arr, columns=columns)
    # df = df.sort_values(by=['IV'], ascending=False)
    df = df.sort_values(by=['IV_sort'], ascending=False)

    # df.to_csv('rule_valid.csv', index=False, header=columns)
    columns.pop()
    df.drop(['IV_sort'], axis=1).to_csv(rule_file_name, index=False, header=columns)
    print('to_csv_success')
    write_data(OutModelFileName, 'to_csv_success')

    if makemodel == 1:
        ks = []
        ks_index = []
        ks_train = []
        ks_index_train = []

        ROC_index = []
        ROC = []
        theta_all = []

        ks_big = []
        ks_index_big = []
        ks_train_big = []
        ks_index_train_big = []

        ROC_index_big = []
        ROC_big = []
        theta_all_big = []

        # initialize data
        for i in range(X_dim):
            if init_flag_list[i] == 1:
                data = ld.rankData(data, i, bond_list[i], woe_list[i])
        # pd.DataFrame(data).to_csv("C:\\Users\\20160822\\Desktop\\Features.csv")

        sum_ks = 0.0
        sum_ks_train = 0.0
        sum_auc = 0.0
        sum_ks_big = 0.0
        sum_ks_train_big = 0.0
        sum_auc_big = 0.0
        is_good = 0
        ii = 0
        jj = 0
        for i in range(test_case):
            print('test case # %d' % (i + 1))
            (trainx, trainy, testx, testy) = ld.getTrainTest(data)
            theta = train.startTrain(trainx, trainy, 0.0)
            (ks_temp, dummy1, dummy2) = perf.KS(testx, testy, theta)
            (ks_temp_train, dummy1_train, dummy2_train) = perf.KS(trainx, trainy, theta)
            if ks_temp_train > ks_temp:
                is_good = 1
                theta_all.append(theta)
                sum_ks += 1.0 * ks_temp
                ks.append((ks_temp, dummy1, dummy2))
                ks_index.append(ks_temp)
                sum_ks_train += 1.0 * ks_temp_train
                ks_train.append((ks_temp_train, dummy1_train, dummy2_train))
                ROC.append(perf.ROC(testx, testy, theta))
                # ROC_index.append(ROC[-1][0])
                sum_auc += 1.0 * ROC[ii][0]
                ii += 1
            else:
                theta_all_big.append(theta)
                sum_ks_big += 1.0 * ks_temp
                ks_big.append((ks_temp, dummy1, dummy2))
                ks_index_big.append(ks_temp)
                sum_ks_train_big += 1.0 * ks_temp_train
                ks_train_big.append((ks_temp_train, dummy1_train, dummy2_train))
                ROC_big.append(perf.ROC(testx, testy, theta))
                # ROC_index_big.append(ROC[-1][0])
                sum_auc_big += 1.0 * ROC_big[jj][0]
                jj += 1

        if is_good == 0:
            theta_all = theta_all_big
            ks = ks_big
            ks_train = ks_train_big
            ROC = ROC_big
            ks_index = ks_index_big
            # ROC_index = ROC_index_big
        ks_index = np.array(ks_index)
        # ROC_index = np.array(ROC_index)
        ks_max_index = np.argmax(ks_index)
        # ROC_max_index = np.argmax(ROC_index)
        print('计算 calcVIF')
        perf.calcVIF(data, columns_dim)


        print_1 = '\"theta\":%s' % theta_all[ks_max_index].tolist()
        print_2 = '\"max_k-s\":%f' % ks[ks_max_index][0]
        print_3 = '\"max_k-s_train\":%f' % ks_train[ks_max_index][0]
        print_4 = '\"max_AUC\":%f' % ROC[ks_max_index][0]
        print_5 = '\"avg_k-s\":%f' % ((sum_ks + sum_ks_big) / test_case)
        print_6 = '\"avg_AUC\":%f' % ((sum_auc + sum_auc_big) / test_case)

        str_out_print = print_1 + '\n' + print_2 + '\n' + print_3 + '\n' + print_4 + '\n' + print_5 + '\n' + print_6
        print (str_out_print)
        # print('python_out \"theta\":%s' % theta_all[ks_max_index].tolist())
        # print (min(theta_all[ks_max_index].tolist()[1:]))
        # print
        # print('python_out \"max_k-s\":%f' % ks[ks_max_index][0])
        # print('python_out \"max_k-s_train\":%f' % ks_train[ks_max_index][0])
        # print('python_out \"max_AUC\":%f' % ROC[ks_max_index][0])

        # print
        # print('python_out \"avg_k-s\":%f' % ((sum_ks + sum_ks_big) / test_case))
        # print('python_out \"avg_AUC\":%f' % ((sum_auc + sum_auc_big) / test_case))
        average_psi, a = perf.PSI(trainx, testx, trainy, testy, theta)
        # perf.KS_draw(ks[ks_max_index][1], ks[ks_max_index][2])
        # perf.ROC_draw(ROC[ks_max_index][1], ROC[ks_max_index][2])

        write_data(OutModelFileName,  str_out_print)
        return perf, ks, ks_train,ROC, ks_max_index,average_psi, a, min(theta_all[ks_max_index].tolist()[1:])
    else:
        return None


# coding:utf-8

import re

'''
切割 模型代码 print 的 结果， 专门用于手动复制的情况 处理得到 画图参数
'''

def get_param_bond_woe_all_theta(str_data):

    params_theta = str_data.split('to_csv_success')
    params = params_theta[0]
    thetas = params_theta[1]

    param_bond = dict()
    woe_all = list()

    params_re_findall = re.findall('python_out.*?\n', params)
    for params_re in params_re_findall:
        param_name = re.findall('"(.*?)"', params_re)[0]
        bond = eval(re.findall('\[.*?\]', params_re)[0])
        # cnt_good = re.findall('\[.*?\]', params_re)[1]
        # cnt_bad = re.findall('\[.*?\]', params_re)[2]
        # bad_rate = re.findall('\[.*?\]', params_re)[3]
        woe = re.findall('\[.*?\]', params_re)[4]
        woe_all.append(eval(woe))

        param_bond[param_name] = [eval(bd) for bd in bond if 'inf' not in bd]
    theta = eval(re.findall('"theta":(.*?)\n', thetas)[0])
    params = list(param_bond.keys())
    bond_list = [param_bond.get(para) for para in params]
    return params, param_bond, woe_all, theta, bond_list

def get_GBK_Name(filename):
    if isinstance(filename, str):
        if filename.count('.') == 1:
            if '.' in filename:
                return filename.split('.')[0] + 'GBK.' + filename.split('.')[1]
    return filename

if __name__ == '__main__':
    # param_bond, woe_all, theta = get_param_bond_woe_all_theta(str_data)
    #
    # print (param_bond)
    # print (woe_all)
    # print (theta)
    print ()
    # print (get_GBK_Name('C:/Users/donglei/Desktop/EsModel/EsNewModel/Csvs/Csv1227/CsvFile/com_param_corr_sheet7.csv'))


# coding:utf-8


from lapras.ReportConfig import score_bond
from lapras.ReportCode.ShowModelResult import show_model_pic
from lapras.ModelCode.main import run_model


'''
运行代码主文件

使用时请注释掉不需要运行的行
'''
def IV(datafilepath):
    if datafilepath is None:
        print("please enter data file path.")
    else:
        run_model(datafilepath,run_type_cus=0)


# coding:utf-8

import re
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

# from DealConfig import csvFile
from lapras.ReportCode.MakeModelDoc import plt_show


def get_params_data_dict(model_name):
    with open(model_name, 'r') as f:
        params_data = f.read()
    f.close()

    params_data_valid_s = re.findall('(python_out.*?)\nAverage', params_data, re.S)
    if params_data_valid_s:
        params_data_valid = params_data_valid_s[-1]
    else:
        print ('model文件有误，请检查文件格式')
        return
    params_data_list = params_data_valid.split('\n')

    params_data_dict = dict()
    for params_data_one in params_data_list:
        if not params_data_one:
            continue
        if 'python_out' not in params_data_one and 'cnt_bad_rate' not in params_data_one:
            continue
        if 'python_out[' in params_data_one:
            continue
        params_data_one = params_data_one.replace('python_out', '').replace(' ', '').replace('"', '')
        param_data_one_list = params_data_one.split('\t')
        param_name = param_data_one_list[0].split(':')[0]
        param_bond = param_data_one_list[0].split(':')[1]
        param_good = param_data_one_list[1].split(':')[1]
        param_bad = param_data_one_list[2].split(':')[1]
        param_bad_rate = param_data_one_list[3].split(':')[1]
        param_IV = param_data_one_list[4].split(':')[1]
        params_data_dict[param_name] = dict()
        params_data_dict[param_name]['bond'] = eval(param_bond)
        params_data_dict[param_name]['good'] = eval(param_good)
        params_data_dict[param_name]['bad'] = eval(param_bad)
        params_data_dict[param_name]['bad_rate'] = eval(param_bad_rate)
        params_data_dict[param_name]['IV'] = eval(param_IV)
    return params_data_dict


def show_singe_param_pic(params_data_dict, param):
    params_data = params_data_dict.get(param)
    if params_data != None:
        y_count = [int(g_i) + int(params_data['bad'][index_g_i]) for index_g_i, g_i in enumerate(params_data['good'])]
        y_rate = [float(br_i) for br_i in params_data['bad_rate']]
        x = list(range(len(y_count)))
        ticks = ['[' + params_data['bond'][index_bd] + ',' + bd + ']' for index_bd, bd in enumerate(params_data['bond'][1:])]

        plt_show(x, ticks,y_count, y_rate, param)


def show_single(model_name, ParamsShow):
    if isinstance(ParamsShow, str):
        ParamsShow = [ParamsShow]

    if not isinstance(ParamsShow, list):
        return
    params_data_dict = get_params_data_dict(model_name)
    for param in ParamsShow:
        # if len(params_data_dict[param]['bond']) < 6:
        #     continue
        print (param)
        show_singe_param_pic(params_data_dict, param)


def show_mutil(model_name):
    params_data_dict = get_params_data_dict(model_name)
    params_data_key_valid = list(params_data_dict.keys())
    # params_data_key_valid = list(params_data_dict.keys())[:800]

    i = 0
    for param in params_data_key_valid:
        # if len(params_data_dict[param]['bond']) < 6:
        #     continue
        print (param)
        show_singe_param_pic(params_data_dict, param)
        i += 1


if __name__ == '__main__':
    # model_name = 'model_20190307_165416'
    # params_data_dict = get_params_data_dict(model_name)
    #
    # param = 'als_m6_id_pdl_orgnum'
    # show_singe_param_pic(params_data_dict, param)

    # params_data_key_valid = list(params_data_dict.keys())
    # # params_data_key_valid = list(params_data_dict.keys())[:800]
    #
    # i = 0
    # for param in params_data_key_valid:
    #     # if len(params_data_dict[param]['bond']) < 6:
    #     #     continue
    #     print (param)
    #     show_singe_param_pic(params_data_dict, param)
    #     i += 1
    pass



# coding:utf-8



'''
计算 模型分 最值，并自动划分区间，为最后展示分数做准备
'''


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from lapras.Settings import col_head,woe_file_output
from lapras.DealConfig import params, A, B, theta, woe_all, cut_points_step, assist_num, bond_list, eps, get_GBK_Name
from lapras.DealConfig import makeModelDataFillnaFileName, scoreCardDataFileName, comParamCorrFileName, checkDataFileName,\
    scoreCardParamFileName, scoreCardPklFileName,woeFileName

import pickle
import math

def get_model_result():
    pass

# 将评分卡分数转换为概率
def score_to_prob(base_score, A=A, B=B):
    odds = math.pow(math.e,(A-base_score)/B)
    prob = odds/(1+odds)
    return prob

def get_score_max_min(A, B, theta, woe_all):
    # 计算当前 A, B, theta, woe_all 下 得分区间 （最大最小值）
    base_score_pri = A - B * theta[0]
    print ('base_score %s'%base_score_pri )
    base_score_max = base_score_pri
    base_score_min = base_score_pri
    for index_theta, theta_one in enumerate(theta):
        if index_theta == 0:
            continue
        base_score_min -= B * theta_one * max(woe_all[index_theta-1])
        base_score_max -= B * theta_one * min(woe_all[index_theta-1])
        a = 1
    print ('base_score_max %s'%base_score_max)
    print ('base_score_min %s' %base_score_min)
    return base_score_min, base_score_max

def get_score_csv(makeModelDataFillnaFileName='make_model_data_fillna_file.csv', scoreCardDataFileName='score_card_data_file.csv', comParamCorrFileName='com_param_corr_file.csv', checkDataFileName='check_data_storeout_file.csv', save_file=True):
    '''
    保存入模变量 变量相似性 得分结果
    :param makeModelDataFillnaFileName:  原始数据文件
    :param scoreCardDataFileName: 入模变量得分
    :param comParamCorrFileName: 入模变量相似度
    :param checkDataFileName: 每个样本得分
    :param save_file: 是否保存文件
    :return:
    '''
    # 读取 原始数据文件 获得入模变量数据
    try:
        df_data = pd.read_csv(makeModelDataFillnaFileName, encoding='utf-8')
    except:
        df_data = pd.read_csv(makeModelDataFillnaFileName, encoding='gbk')
    params_list = col_head + params
    df_params = df_data[params_list]

    woe_list = []

    # 整理 bond
    bond_all = list()
    for bond in bond_list:
        bond_all.append(['-inf'] + bond + ['inf'])
    data = df_params.drop([col_head[1]],axis=1).values.tolist()
    datay = df_params[col_head[1]].values.tolist()
    dim = len(bond_all)
    fo_list = list()
    for i in range(len(data)):
        cus_no = data[i][0]
        woe_list_tmp = [cus_no,df_params.iloc[i, 1]]

        base_score = A - B * theta[0]
        # print(base_score)
        for j in range(dim):
            cell = data[i][j + 1]
            bond = bond_all[j]
            woe = woe_all[j]
            woe_size = len(woe)
            index = -1
            for k in range(woe_size):
                # print(str(cus_no) + " " + str(bond))
                if k == 0 and cell < bond[k + 1]:
                    index = k
                if k > 0 and k < woe_size - 1 and cell > bond[k] - eps and cell < bond[k + 1]:
                    index = k
                if k == woe_size - 1 and cell > bond[k] - eps:
                    index = k
            # print (B * theta[j + 1] * woe[index])
            base_score -= (B * theta[j + 1] * woe[index])
            woe_list_tmp.append(woe[index])
        # print(base_score)
        fo_list.append([cus_no, datay[i], base_score, score_to_prob(base_score)])
        woe_list_tmp.append(base_score)
        woe_list.append(woe_list_tmp)



    #生成评分卡参数文件
    score_card_map = {}
    column = ['variable', 'bin', 'points']
    intercept = round(A - B * theta[0],2)
    df = pd.DataFrame(['basepoints' ,np.NAN, intercept]).T
    df.columns = column
    score_card_map['basepoints'] = df
    for j in range(len(bond_all)):
        bond = bond_all[j]
        woe = woe_all[j]
        woe_size = len(woe)
        index = -1
        df1 = pd.DataFrame(columns=column)
        for k in range(woe_size):
            if k == 0:
                bin_str = "[" + str(bond[k]) + "," + str(bond[k+1]) + ")"
            else:
                bin_str = "[" + str(bond[k]) + "," + str(bond[k + 1]) + ")"
            score = round(-(B * theta[j + 1] * woe[k]),2)
            df2 = pd.DataFrame([params[j],bin_str,score]).T
            df2.columns = column
            df1 = pd.concat([df1,df2],ignore_index=True)
        score_card_map[params[j]] = df1

    # 将数据序列化后存储到文件中
    f = open(scoreCardPklFileName, 'wb')  # pickle只能以二进制格式存储数据到文件
    f.write(pickle.dumps(score_card_map))  # dumps序列化源数据后写入文件
    f.close()
    f = open(scoreCardParamFileName, 'w')  # 写入文本文件
    f.write(str(score_card_map))
    f.close()


    print ('save_file')
    if save_file:
        # 保存入模变量
        try:
            df_params.to_csv(scoreCardDataFileName, index=None, encoding='utf-8')
            # df_params.to_csv(get_GBK_Name(scoreCardDataFileName), index=None, encoding='gbk')
            print ('save scoreCardDataFileName success')
        except:
            print('error in save scoreCardDataFileName')
        # 保存 变量相似性
        df_data[params].corr().to_csv( comParamCorrFileName, index=None, encoding='utf-8', )
        print('save comParamCorrFileName success')
        if woe_file_output:
            pd.DataFrame(woe_list, columns=params_list + ['base_score']).to_csv( woeFileName, index=None, encoding='utf-8')
            print('save woe file success')

        # 保存得分结果
        # df_fo = pd.DataFrame(fo_list, columns=['cus_no', 'y', 'base_score'])
        df_fo = pd.DataFrame(fo_list, columns=col_head + ['base_score','default_prob'])
        try:
            df_fo.to_csv(checkDataFileName, index=False, encoding='utf-8')
            # df_fo.to_csv(get_GBK_Name(checkDataFileName), index=False, encoding='gbk')
            print ('save checkDataFileName success')
        except:
            print('error in save checkDataFileName')
    return

def get_score_bond(cut_points_step, assist_num=assist_num, minpoint=200, maxpoint=1000,  checkDataFileName='check_data_file.csv'):
    try:
        scoredata = pd.read_csv(checkDataFileName, encoding='utf-8')
    except:
        scoredata = pd.read_csv(checkDataFileName, encoding='gbk')

    base_score = scoredata['base_score']
    base_score_max = max(base_score)
    base_score_min = min(base_score)
    maxpoint = int(maxpoint) + 1
    minpoint = int(minpoint) - 1
    cutmin = int(base_score_min / 10.0) * 10
    cutmax = int((base_score_max / 10.0) * 10)
    while (scoredata[(scoredata['base_score'] <= cutmin) & (scoredata[col_head[1]] == 0)]['base_score'].count() < 20 or scoredata[(scoredata['base_score'] <= cutmin) & (scoredata[col_head[1]] == 1)]['base_score'].count() < 1):
        cutmin = cutmin + 10
    while (scoredata[(scoredata['base_score'] >= cutmax) & (scoredata[col_head[1]] == 0)]['base_score'].count() < 20 or scoredata[(scoredata['base_score'] >= cutmax) & (scoredata[col_head[1]] == 1)]['base_score'].count() < 1):
        cutmax = cutmax - 10
    cut_points = list(range(cutmin, cutmax + assist_num * cut_points_step, cut_points_step))
    break_points = [minpoint] + cut_points + [maxpoint]
    return break_points


if __name__ == '__main__':

    # 保存入模变量 变量相似性 得分结果
    get_score_csv(makeModelDataFillnaFileName, scoreCardDataFileName, comParamCorrFileName, checkDataFileName, save_file=False)
    # 得到计分区间
    base_score_min, base_score_max = get_score_max_min(A, B, theta, woe_all)
    # 自动划分区间
    break_points = get_score_bond(cut_points_step, assist_num=assist_num, minpoint=base_score_min, maxpoint=base_score_max)
    print (break_points)
